/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "common.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>

#include "child.h"
#include "nntp.h"

static ChildData child_data = CHILD_NULL;

static BOOL child_shouldnt_exit = TRUE;

static void child_exit(void)
{
    if (child_shouldnt_exit) 
    {
	fprintf(stderr,
		_("<%d> Warning: Child process exiting unexpectedly\n"),
		child_data.thread_idx);
    }
}

void child_send_command(intercom_Command code, const char *data, ...)
{
    va_list ap;

    va_start(ap, data);
    intercom_send_vcommand(child_data.out_buf, code, child_data.thread_idx,
	    data, ap);
    va_end(ap);
}

static BOOL child_sent_quit = FALSE;

static void child_shutdown(BOOL error)
{
    /* Try to let server know we're leaving as gracefully as possible, but
     * don't hang around for response or anything if there's an error */
    deblog_info(_("Thread shutting down %s\n"), error ? _("due to error") :
	    _("naturally"));
    if (child_data.sbuf)
    {
	if (!child_sent_quit)
	{
	    child_sent_quit = TRUE;
	    rw_write_line(child_data.sbuf, "QUIT");
	}
	if (!error)
	{
	    nntp_check_response(child_data.sbuf, child_data.server,
		    _("Bad response to QUIT"), "2", deblog_Level_Critical);
	}
	/* Reset in case this thread services another server */
	child_sent_quit = FALSE;
	rw_close_and_free(child_data.sbuf);
	child_data.sbuf = NULL;
	deblog_info(_("Closed socket OK\n"));
    }
    if (child_data.intercom)
    {
	child_send_command(error ? command_QuietError : command_AllDone,
		"");
    }
    free(child_data.server);
    child_data.server = NULL;
    child_shouldnt_exit = FALSE;
}

static void child_signal_handler(int sig)
{
    static BOOL recursive_exit = FALSE;
    
    if (!recursive_exit)
    {
	recursive_exit = TRUE;
	if (!child_sent_quit)
	{
	    child_sent_quit = TRUE;
	    if (child_data.sbuf)
		rw_write_line(child_data.sbuf, "QUIT");
	}
	fprintf(stderr, _("Child exiting due to signal %d\n"), sig);
	fflush(stderr);
	exit(1);
    }
    else
	abort();
}

static void child_connect(const char *comdata)
{
    char dotquad[32];
    int port;
    char *path = XMALLOC(strlen(comdata));
    char *servid = XMALLOC(strlen(comdata));
    BOOL ssl;

    child_shouldnt_exit = TRUE;
    child_data.pipeline_depth = 0;
    sscanf(comdata, "%d %s %s %s %d %d %d", &port, servid, dotquad, path,
	    &child_data.pipeline, &child_data.ignore_436, &ssl);
    child_data.server = servid;
    deblog_debug(_("Connecting to %s, pipelining to depth %d %s\n"), servid,
	    PIPELINE_DEPTH, child_data.pipeline ? _("enabled") : _("disabled"));
    if (child_data.ignore_436)
    {
	deblog_debug(_("Ignoring 441 436 <Message-ID> POST rejections\n"));
    }
    child_data.sbuf = nntp_connect(child_data.server, dotquad,
	    port, child_data.thread_idx, ssl);
    if (!child_data.sbuf)
    {
	child_shutdown(TRUE);
	return;
    }
    child_send_command(command_Connected, "");
    free(child_data.path);
    if (strcmp(path, "!"))
    {
	child_data.path = path;
    }
    else
    {
	free(path);
	child_data.path = NULL;
    }
}

static void child_authenticate(const char *usrnpwd)
{
    char *user, *passwd;

    user = XMALLOC(strlen(usrnpwd) + 1);
    passwd = XMALLOC(strlen(usrnpwd) + 1);
    sscanf(usrnpwd, "%s %s", user, passwd);
    if (nntp_authenticate(child_data.sbuf, child_data.server, user, passwd))
    {
	child_shutdown(TRUE);
    }
    else
    {
	child_send_command(command_Authenticated, "");
    }
    free(user);
    free(passwd);
}

static void child_request_article(void)
{
    child_send_command(command_GetNextArticle, "%ld",
	    child_data.rq_group_ix);
}

static void child_get_group_stats(const char *data)
{
    char *groupname = XMALLOC(strlen(data) + 1);
    long int first, last;

    sscanf(data, "%ld %s", &child_data.rq_group_ix, groupname);
    deblog_debug(_("Entering group %ld %s to get stats\n"),
	    child_data.rq_group_ix, groupname);
    if (nntp_enter_group(child_data.sbuf, child_data.server, groupname,
		&first, &last, TRUE))
    {
	child_send_command(command_BogusGroup, "%ld",
		child_data.rq_group_ix);
    }
    else
    {
	deblog_debug(_("%s: %ld-%ld\n"), groupname, first, last);
	child_send_command(command_GroupArticleStats, "%ld %ld %ld",
		child_data.rq_group_ix, first, last);
    }
    free(groupname);
}

static void child_issue_xhdr(const char *data, BOOL xover)
{
    const char *line;
    long int group_ix;
    char *range = XMALLOC(strlen(data));

    sscanf(data, "%ld %s", &group_ix, range);
    rw_writef(child_data.sbuf, "%s %s\r\n",
	    xover ? "XOVER" : "XHDR Message-ID", range);
    deblog_debug("Issuing %s %s\r\n",
	    xover ? "XOVER" : "XHDR Message-ID", range);
    free(range);
    if (!(line = nntp_check_response(child_data.sbuf, child_data.server,
		_("Bad response to XHDR/XOVER"), "2", deblog_Level_Critical)))
    {
	child_shutdown(TRUE);
	return;
    }
    deblog_debug(_("Response to XOVER/XHDR command: %s\n"), line);
    while ((line = rw_read_line(child_data.sbuf)) != NULL && line[0] != '.')
    {
	deblog_debug(_("XOVER/XHDR line: %s\n"), line);
	child_send_command(xover ? command_XoverData : command_GroupContains,
		"%ld %s", group_ix, line);
    }
    child_send_command(command_GroupComplete, "%ld", group_ix);
}

static const char *child_path_filter(const char *line, void *data)
{
    static char *newline = NULL;
    static size_t nllen = 0;
    size_t len;

    UNUSED(data);

    if (child_data.path_done)
	return line;
    if (!line[0])
    {
	child_data.path_done = TRUE;
	return line;
    }
    if (strncmp(line, "Path: ", 6))
	return line;
    len = strlen(line) + strlen(child_data.path) + 2;
    if (nllen < len)
	newline = XREALLOC(newline, nllen = len);
    sprintf(newline, "Path: %s!%s", child_data.path, line + 6);
    child_data.path_done = TRUE;
    return newline;
}

static void child_shift_pipeline(void)
{
    int i;

    for (i = 1; i < child_data.pipeline_depth; ++i)
    {
	child_data.group_ix[i - 1] = child_data.group_ix[i];
	child_data.artnum[i - 1] = child_data.artnum[i];
	child_data.change_group[i - 1] = child_data.change_group[i];
	child_data.backfill[i - 1] = child_data.backfill[i];
    }
    --child_data.pipeline_depth;
}

/* Returns TRUE if problem */
static BOOL child_pop_article(void)
{
    BOOL result = FALSE;

    if (child_data.change_group[0])
    {
	deblog_info(
		_("Child now expecting to receive articles from group %ld\n"),
		child_data.group_ix[0]);
	if (!nntp_check_response(child_data.sbuf, child_data.server,
		    _("Can't enter group"), "2", deblog_Level_Critical))
	{
	    child_shutdown(TRUE);
	    return TRUE;
	}
	child_send_command(command_EnteredGroup, "%ld",
		child_data.group_ix[0]);
    }
    else
    {
	int fd = -1;
	rw_buf *fbuf = NULL;
	char *filename = NULL;

	child_data.path_done = FALSE;
	deblog_debug(_("Expecting to receive %d/%d\n"), child_data.group_ix[0],
		    child_data.artnum[0]);
	if (!nntp_check_response(child_data.sbuf, child_data.server,
		    _("Bad response to ARTICLE"), "2",
		    child_data.backfill[0] ?
		    deblog_Level_Info : deblog_Level_Warning))
	{
	    /* Don't need to report back if it's a backfill */
	    if (!child_data.backfill[0])
	    {
		child_send_command(command_ArticleUnavailable, "%d %ld",
			child_data.group_ix[0], child_data.artnum[0]);
	    }
	    result = TRUE;
	}
	if (!result)
	{
	    fd = xmktmpindir(INCOMING_DIR, "Mesg", &filename, FALSE,
		    _("Can't create temporary file to download message"));
	    if (fd < 0)
	    {
		child_shutdown(TRUE);
		result = TRUE;
	    }
	    else
	    {
		deblog_debug(_("Saving %d/%d to %s\n"),
			child_data.group_ix[0], child_data.artnum[0],
			filename);
		fbuf = rw_create_buf(fd, FALSE, FALSE, -1);
		if (rw_stof(child_data.sbuf, fbuf, child_data.path ?
			    child_path_filter : NULL, NULL) < 0)
		{
		    deblog_critical(
			    _("Can't download message to temporary file\n"));
		    child_shutdown(TRUE);
		    result = TRUE;
		}
	    }
	    if (fbuf)
		rw_free_buf(fbuf);
	    if (fd != -1)
		close(fd);
	    if (!result)
	    {
		child_send_command(command_HaveDownloaded, "%ld %ld %d %s",
			child_data.group_ix[0], child_data.artnum[0],
			child_data.backfill[0], filename);
	    }
	}
	free(filename);
    }
    child_shift_pipeline();
    return result;
}

static void child_get_article(const char *data)
{
    sscanf(data, "%ld %ld %d",
	    &child_data.group_ix[child_data.pipeline_depth],
	    &child_data.artnum[child_data.pipeline_depth],
	    &child_data.backfill[child_data.pipeline_depth]);
    child_data.change_group[child_data.pipeline_depth] = FALSE;
    rw_writef(child_data.sbuf, "ARTICLE %ld\r\n",
	    child_data.artnum[child_data.pipeline_depth]);
    ++child_data.pipeline_depth;
    deblog_debug(_("Increasing pipeline depth to %d\n"),
	    child_data.pipeline_depth); 
    if (child_data.pipeline_depth >= (child_data.pipeline ? PIPELINE_DEPTH : 1))
    {
	deblog_debug(_("Pipeline full, popping\n"));
	child_pop_article();
    }
    child_request_article();
}

static void child_enter_group(const char *data)
{
    char *groupname = XMALLOC(strlen(data) + 1);

    sscanf(data, "%ld %s", &child_data.rq_group_ix, groupname);
    deblog_debug(_("Entering group %ld %s to get articles\n"),
	    child_data.rq_group_ix, groupname);
    child_data.group_ix[child_data.pipeline_depth] = child_data.rq_group_ix;
    child_data.change_group[child_data.pipeline_depth] = TRUE;
    nntp_enter_group(child_data.sbuf, child_data.server, groupname,
	    NULL, NULL, FALSE);
    ++child_data.pipeline_depth;
    deblog_debug(_("Increasing pipeline depth to %d\n"),
	    child_data.pipeline_depth); 
    if (child_data.pipeline_depth >= (child_data.pipeline ? PIPELINE_DEPTH : 1))
    {
	deblog_debug(_("Pipeline full, popping\n"));
	child_pop_article();
    }
    deblog_debug(_("Entered %s, getting article\n"), groupname);
    child_request_article();
    free(groupname);
}

static void child_resolve(const char *data)
{
    char *dotquad = nntp_resolve_address(data);

    if (dotquad)
	child_send_command(command_Resolved, dotquad);
    else
	child_shutdown(TRUE);
    free(dotquad);
}

static void whinge_unlink(const char *pathname, int e)
{
    deblog_critical(_("%s posted OK (or was duplicate),\n"), pathname);
    deblog_critical(_("but couldn't be removed from outgoing queue:\n"));
    deblog_critical("%s\n", strerror(e));
}

static void child_post_article(const char *pathname)
{
    rw_buf *fbuf;
    int fd;
    const char *response = NULL;
    int result;
    BOOL close_busy = FALSE;

    /* Only need to read file, but has to be opened writable for lockf */
    if ((fd = open(pathname, O_RDWR)) == -1)
    {
	if (errno != ENOENT)
	{
	    int e = errno;
	    char *reason = XSTRDUPCAT(_("Couldn't open for uploading: "),
		    strerror(e));

	    deblog_critical(_("Couldn't open %s for uploading:\n"), pathname);
	    deblog_critical("%s\n", strerror(e));
	    child_send_command(command_BadUpload, "%s %s %s",
		pathname, child_data.server, reason);
	}
	else
	{
	    deblog_info(
		    _("%s doesn't exist for uploading (already uploaded)\n"),
		    pathname);
	    child_send_command(command_LocalDuplicate, "");
	}
	return;
    }
    if (lockf(fd, F_TLOCK, 0))
    {
	int e = errno;

	if (e == EACCES || e == EAGAIN)
	{
	    deblog_info(_("%s locked for uploading (already being uploaded)\n"),
		    pathname);
	    child_send_command(command_LockedDuplicate, "");
	}
	else
	{
	    char *reason = XSTRDUPCAT(_("Couldn't lock for uploading: "),
		    strerror(e));

	    child_send_command(command_BadUpload, "%s %s %s", pathname,
		    child_data.server, reason);
	    free(reason);
	}
	close(fd);
	return;
    }
    deblog_info(_("Uploading %s\n"), pathname);
    fbuf = rw_create_buf(fd, FALSE, FALSE, -1);
    rw_write_line(child_data.sbuf, "POST");
    if (!nntp_check_response(child_data.sbuf, child_data.server,
		_("Bad response to POST"), "3", deblog_Level_Critical))
    {
	rw_close_and_free(fbuf);
	child_send_command(command_NoUploads, "");
	return;
    }
    result = rw_ftos(fbuf, child_data.sbuf, NULL, NULL);
    if (!result)
    /* From now on result == 0 means posting was successful, or rejected
     * because it's a duplicate, and response is reason for failure */
    {
	response = rw_read_line(child_data.sbuf);
	deblog_debug("%s\n", response ? response : _("No response"));
	if (!response)
	{
	    child_send_command(command_Error,
		_("No response from server when posting\n"));
	    return;
	}
	else if (!maincf_get_report_dups() &&
		    (!strncmp(response, "441 435 ", 8)
		    || strstr(response, "Duplicate")
		    || strstr(response, "Already got")
		    || (child_data.ignore_436 &&
			!strncmp(response, "441 436 <", 9))))
	{
	    deblog_info(_("Upload %d already known by server\n"), pathname);
	    child_send_command(command_RemoteDuplicate, pathname);
	}
	else if (response[0] != '2')
	{
	    result = -1;
	}
	else
	{
	    deblog_info(_("%s uploaded successfully\n"), pathname);
	    child_send_command(command_Uploaded, pathname);
	}
    }
    else
    {
	response = strerror(errno);
    }
    /* Lock means we should remove file before closing it, but that could be
     * impossible on some systems */
    if (result)
    {
	deblog_info(_("Posting to %s failed: %s\n"),
		child_data.server, response);
	child_send_command(command_BadUpload, "%s %s %s",
		pathname, child_data.server, response);
    }
    else
    {
	if (unlink(pathname))
	{
	    if (errno == EBUSY)
	    {
		close_busy = TRUE;
		deblog_info(_("Can't delete %s after uploading "
			    "while same thread still has a lock.\n"),
			pathname);
		deblog_info(_("This is a system limitation. "
			    "Will retry after releasing lock.\n"));
	    }
	    else
	    {
		whinge_unlink(pathname, errno);
	    }
	}
    }
    rw_close_and_free(fbuf);
    if (close_busy && unlink(pathname) && errno != ENOENT)
	whinge_unlink(pathname, errno);
}

static void child_flush_pipeline(void)
{
    deblog_info(_("Flushing pipelined articles prior to next task\n"));
    while (child_data.pipeline_depth)
	child_pop_article();
    if (child_data.all_requested)
    {
	deblog_message(_("All articles downloaded\n"));
	child_data.all_requested = FALSE;
	child_send_command(command_AllDownloaded, "");
    }
}

static void child_find_xref_in_xover(void)
{
    int i = 0;
    const char *line;
    int ix = -1;
    BOOL includes_name = FALSE;
    BOOL is_newsgroups = FALSE;

    rw_write_line(child_data.sbuf, "LIST OVERVIEW.FMT");
    if (!nntp_check_response(child_data.sbuf, child_data.server,
		_("Can't get overview format"), "2", deblog_Level_Warning))
    {
	child_send_command(command_FoundXrefInXover, "-1 0 0");
	return;
    }
    for (i = 0; (line = rw_read_line(child_data.sbuf)) != NULL 
	    && strcmp(line, "."); ++i)
    {
	if (!strncmp(line, "Newsgroups:", 11))
	{
	    is_newsgroups = TRUE;
	}
	else if (ix == -1 && !strncmp(line, "Xref:", 5))
	{
	    is_newsgroups = FALSE;
	}
	else
	{
	    /* Skip rest of loop body if not a match */
	    continue;
	}
	ix = i;
	deblog_debug(_("Found %s in overview.fmt at ix %d\n"),
		line, ix);
	/* See if line ends in "full" */
	if ((line = strstr(line, "full")) != NULL && !strcmp(line, "full"))
	{
	    deblog_debug(
		    _("Overview for Newsgroups/Xref includes header name\n"));
	    includes_name = TRUE;
	}
	else
	{
	    deblog_debug(
		    _("Overview for Newsgroups/Xref excludes header name\n"));
	}
    }
    child_send_command(command_FoundXrefInXover, "%d %d %d", ix,
	    includes_name, is_newsgroups);
}

static void child_main_loop(void)
{
    const intercom_data *command;

    for (;;)
    {
	command = intercom_get_command(child_data.intercom);
	switch (command->command)
	{
	    case command_Upload:
		child_flush_pipeline();
		child_post_article(command->data);
		break;
	    case command_Resolve:
		child_resolve(command->data);
		break;
	    case command_Connect:
		child_connect(command->data);
		break;
	    case command_ModeReader:
		if (nntp_set_mode(child_data.sbuf, child_data.server, "READER"))
		    child_shutdown(TRUE);
		else
		    child_send_command(command_Ready, "");
		break;
	    case command_Auth:
		child_authenticate(command->data);
		break;
	    case command_FindXrefInXover:
		child_find_xref_in_xover();
		break;
	    case command_GetGroupStats:
		child_flush_pipeline();
		child_get_group_stats(command->data);
		break;
	    case command_EnterGroup:
		child_enter_group(command->data);
		break;
	    case command_GetArticles:
		child_issue_xhdr(command->data, FALSE);
		break;
	    case command_GetXover:
		child_issue_xhdr(command->data, TRUE);
		break;
	    case command_GetArticle:
		child_get_article(command->data);
		break;
	    case command_AllRequested:
		child_data.all_requested = TRUE;
		break;
	    case command_Throttle:
		usleep(atoi(command->data) * 1000);
		child_request_article();
		break;
	    case command_Shutdown:
		deblog_info(
		    _("<%d> Received shutdown command, %d %s in pipeline\n"),
		    child_data.thread_idx,
		    child_data.pipeline_depth,
		    child_data.pipeline_depth == 1 ? _("article") :
			_("articles"));
		child_sent_quit = TRUE;
		if (child_data.sbuf)
		    rw_write_line(child_data.sbuf, "QUIT");
		child_flush_pipeline();
		child_shutdown(FALSE);
		break;
	    case command_WaitPending:
		child_flush_pipeline();
		child_send_command(command_WaitAck, "");
		break;
	    default:
		deblog_error(_("Child doesn't understand command code %d\n"),
			command->command);
		break;
	}
    }
}

void child_start(int thread_idx, int in_fd, int out_fd)
{
    static BOOL exit_handled = FALSE;

    if (!exit_handled)
    {
	atexit(child_exit);
	exit_handled = TRUE;
    }
    signal(SIGHUP, child_signal_handler);
    signal(SIGINT, SIG_IGN);
    signal(SIGPIPE, child_signal_handler);
    signal(SIGTERM, child_signal_handler);
    signal(SIGALRM, SIG_IGN);
    signal(SIGUSR1, SIG_IGN);
    signal(SIGUSR2, SIG_IGN);
    child_shouldnt_exit = FALSE;
    child_data.in_buf = rw_create_buf(in_fd, FALSE, TRUE, -1);
    child_data.out_buf = rw_create_buf(out_fd, FALSE, TRUE, -1);
    child_data.intercom = intercom_new(child_data.in_buf);
    child_data.thread_idx = thread_idx;
    child_data.pipeline_depth = 0;
    child_data.path = NULL;
    deblog_use_pipe(child_data.out_buf, thread_idx);
    child_main_loop();
}

