/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <limits.h>

#include "intercom.h"

Intercom *intercom_new(rw_buf *buf)
{
    Intercom *ic = XMALLOC(sizeof(Intercom));

    ic->buf = buf;
    ic->data.data = ic->str_data = NULL;
    return ic;
}

void intercom_delete(Intercom *ic)
{
    if (!ic)
	return;
    free(ic->str_data);
    free(ic);
}

void intercom_send_command(rw_buf *buf, intercom_Command code, int thread_idx,
	const char *data, ...)
{
    va_list ap;
    
    va_start(ap, data);
    intercom_send_vcommand(buf, code, thread_idx, data, ap);
    va_end(ap);
}

#define INTERCOM_ATOMIC PIPE_BUF
/* NB configure script checks that PIPE_BUF is large enough to be usable */

#define MAX_RECURSION 500

/* This makes recursion for long commands more efficient */
static void intercom_send_formatted_command(rw_buf *buf, int thread_idx,
	const char *headstr, const char *formstr, int recursion)
{
    static BOOL baled_out = FALSE;

    if (recursion > MAX_RECURSION)
    {
	fprintf(stderr,
		"\nToo much recursion in intercom_send_formatted_command, ");
	if (!baled_out)
	{
	    baled_out = TRUE;
	    fprintf(stderr, "exiting\n");
	    fflush(stderr);
	    exit(1);
	}
	else
	{
	    fprintf(stderr, "aborting\n");
	    fflush(stderr);
	    abort();
	}
    }
    /* +2 is for CRLF terminator - should only need LF but play safe */
    if (thread_idx != -1 &&
	    strlen(headstr) + strlen(formstr) + 2 >= INTERCOM_ATOMIC)
    {
	char splitheadstr[8];
	/* malloc in case it isn't a good idea to have something so large on
	 * stack */
	char *chopped = XMALLOC(INTERCOM_ATOMIC);
	size_t frag_len;

	/*
	printf("\nThread %d split:\n"
		"headstr (%d) + formstr (%d) + 2 = "
		"%d >= INTERCOM_ATOMIC (%d)\n",
		strlen(headstr), strlen(formstr),
		strlen(headstr) + strlen(formstr) + 2,
		INTERCOM_ATOMIC);
	*/
	snprintf(splitheadstr, 8, "%d %c ", thread_idx,
		command_SplitCommand);
	frag_len = INTERCOM_ATOMIC - 3 - strlen(splitheadstr);
	strncpy(chopped, formstr, frag_len);
	chopped[frag_len] = 0;
	/* Recursion, woo! */
	/*
	printf("Splitting %s %s\ninto %s %s\n%s %s\n",
		headstr, formstr,
		splitheadstr, chopped,
		headstr, formstr + frag_len);
	sleep(1);
	*/
	intercom_send_formatted_command(buf, thread_idx,
		splitheadstr, chopped, recursion + 1);
	intercom_send_formatted_command(buf, thread_idx,
		headstr, formstr + frag_len, recursion + 1);
    }
    else
    {
	char *longstr = XSTRDUPCAT(headstr, formstr);

	if (rw_write_line(buf, longstr) <= 0)
	{
	    fprintf(stderr, _("Can't send interprocess command: %s\n %s\n"),
		    strerror(errno), longstr);
	    /*
	    perror(_("Can't send interprocess command"));
	    */
	    exit(1);
	}
	free(longstr);
    }
}

void intercom_send_vcommand(rw_buf *buf, intercom_Command code, int thread_idx,
	const char *data, va_list ap)
{
    char *formstr;
    char headstr[8];

    if (vasprintf(&formstr, data, ap) < 0)
    {
	fprintf(stderr, 
		_("Can't send interprocess command (out of memory?)\n"));
	exit(1);
    }
    snprintf(headstr, 8, "%d %c ", thread_idx, code);
    intercom_send_formatted_command(buf, thread_idx, headstr, formstr, 0);
    free(formstr);
}

const intercom_data *intercom_get_command(Intercom *ic)
{
    const char *line;
    size_t len;
    
    free(ic->str_data);
    ic->str_data = NULL;
    line = rw_read_line(ic->buf);
    if (!line)
    {
	deblog_error(_("Can't read interprocess command: %s\n"),
		strerror(errno));
	return NULL;
    }
    ic->data.command = command_Null;
    if (sscanf(line, "%d %c", &ic->data.thread_idx, (char *)
		&ic->data.command) < 2
	    || ic->data.command < command_First
	    || ic->data.command > command_Last)
    {
	deblog_error(_("Received bad IPC command line: %s\n"), line);
	free(ic->str_data);
	ic->str_data = NULL;
	return NULL;
    }
    ic->str_data = XSTRDUP(strchr(strchr(line, ' ') + 1, ' ') + 1);
    /* Log message may have been sent with \n replaced by \r, so switch back */
    len = strlen(ic->str_data) - 1;
    if (ic->str_data[len] == '\r')
	ic->str_data[len] = '\n';
    ic->data.data = ic->str_data;
    /*
    if (ic->data.command != command_Log)
    {
	deblog_debug(_("Received IPC line: %s\n"), line);
    }
    */
    return &ic->data;
}

