#ifndef INTERCOM_H
#define INTERCOM_H

/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* For exchange of commands between processes via pipes */


#ifndef COMMON_H
#include "common.h"
#endif

/* Each command consists of space separated fields:
 *
 * Thread idx (can be anything for parent, because children don't listen to
 * anything else)
 * 
 * Command code
 * 
 * Data depending on command
 */

typedef enum {
    command_Null = 0,
    command_First = 'A',
    /* From child to parent */
    command_Ready = command_First,	/* No data, logged on ready to receive
					   first command */
    command_Resolved,			/* IP address (dotted-quad) requested
					   by command_Resolve */
    command_Connected,
    command_Authenticated,
    command_FoundXrefInXover,		/* First arg is field index of Xref;
					   2nd arg bool: field includes header
					   name; 3rd arg bool: this is actual
					   newsgroups header, not Xref */
    command_GroupArticleStats,		/* Group idx, first, last art numbers */
    command_GroupContains,		/* Group idx, article no, Message-ID */
    command_XoverData,			/* Group idx, XOVER data for one
					   article */
    command_GroupComplete,		/* Group idx:- all IDs read */
    command_GetNextArticle,		/* Group idx: request next message in
					   given group */
    command_HaveDownloaded,		/* Group idx, article no, backfill
					   (bool), filename */

    command_Uploaded,			/* Arg: pathname */
    command_RemoteDuplicate,		/* Arg: pathname */		
    command_LockedDuplicate,
    command_LocalDuplicate,
    command_BadUpload,			/* Args: pathname, server, reason */

    command_NoUploads,			/* If posting fails because of the
					   message, it's child's responsibility
					   to move it to "failed" directory. If
					   posting fails in general, it sends
					   this message instead, and parent
					   should skip uploading rest of
					   server's outgoing articles. */
    command_Log,			/* Level, message */
    command_AmountRcvd,			/* Bytes received from server since arg
					   2; timeval */
    command_AmountSent,			/* As above, but bytes sent */
    command_Error,			/* Critical error with message to be
					   logged; implies command_AllDone */
    command_QuietError,			/* As error, without a message */
    command_AllDone,			/* Usually in response to
					   command_Shutdown */
    command_BogusGroup,			/* Group idx: not a valid group
					   according to server */
    command_ArticleUnavailable,		/* Group idx, article no */
    command_EnteredGroup,		/* Group ix: Just entered this group
					   for fetching */
    command_AllDownloaded,		/* All articles have been downloaded,
					   time to show message */
    command_WaitAck,			/* Pipeline has been flushed, ready
					   to enter wait state */
    command_SplitCommand,		/* Command has been split because it's
					   longer than PIPE_BUF. Data should be
					   saved and prepended to following
					   command, which will either be
					   another command_SplitCommand with
					   middle section of data or some other
					   command with the tail of the data */

    /* From parent to child */
    command_Resolve,			/* Address (name) to be resolved */
    command_Connect,			/* Port; server id; server address
					   (given in dotted-quad); entry to add
					   to Path header or !; whether to
					   pipeline (0/1); whether to ignore
					   441 436 POST rejections; whether to
					   use SSL
					 */
    command_FindXrefInXover,		/* LIST OVERVIEW.FMT to find whether it
					   contains Xref (or Newsgroups) field
					 */
    command_GetGroupStats,		/* Group ix, name: Enter group and
					   reply with command_GroupContains */
    command_EnterGroup,			/* Group ix, name: Enter group, ready
					   to fetch articles from it */
    command_Auth,			/* User name, password */
    command_LogOn = command_Auth,
    command_ModeReader,			/* Send MODE READER command */
    command_GetArticles,		/* Send series of GroupContains;
					   group ix, first, last */
    command_GetXover,			/* Send series of XoverData;
					   group ix, first, last */
    command_GetArticle,			/* Group ix, server article no,
					   backfill (bool) */
    command_AllRequested,		/* All articles have been requested, so
					   when pipeline is flushed, show
					   "Download complete" message */
    command_Upload,			/* Full filename */
    command_Throttle,			/* Caught up with list of IDs, or using
					   too much bandwidth, pause for given
					   time (ms) */
    command_WaitPending,		/* Tell it to send
					   command_WaitAck when
					   pipeline is flushed */
    command_Shutdown,			/* Pull any messages remaining in
					   queue, then log off server and stop
					 */

    command_Last = command_Shutdown
} intercom_Command;

struct intercom_data {
    intercom_Command command;
    int thread_idx;
    const char *data;
};

struct Intercom {
    rw_buf *buf;
    intercom_data data;
    char *str_data;		/* Needed because data (string) in
				   intercom_data is const */
};

Intercom *intercom_new(rw_buf *);

void intercom_delete(Intercom *);

void intercom_send_command(rw_buf *, intercom_Command, int thread_idx, const
	char *data, ...);

void intercom_send_vcommand(rw_buf *, intercom_Command, int thread_idx, const
	char *data, va_list);

const intercom_data *intercom_get_command(Intercom *);

#endif /* INTERCOM_H */
